Program RDBF;

{*****************************************************************************
08/06/88  TJW  .DBF accessing routines are from Micro Cornucopia Sept/Oct 1988

Compile with Turbo Pascal version 3.02.  (May not work with version 4.)
Purpose: Displays, Prints or Creates file of dBase .DBF file structure.

Usage:
         RDBF             - Asks for file name.
         RDBF file        - Displays the file definition.
         RDBF file.dbf    - Same as above.
         RDBF file f      - Displays and creates "file.DEF".
         RDBF file.dbf f  - Same as above.
         RDBF file p      - Displays and prints the file definition.
         RDBF file.dbf p  - Same as above.

*****************************************************************************}

Type

  dbf_hdr_rec = record   {header structure record}
    headtype    : byte;  {control byte - 03h = no memo file, 83h = memo file}
    year        : byte;  {last access}
    month       : byte;  {last access}
    day         : byte;  {last access}
    rec_count1  : integer; {stored as unsigned integer}
    rec_count2  : integer; {stored as unsigned integer}
    hdr_length  : integer; {in bytes}
    rec_size    : integer; {in bytes}
  end;

  dbf_fld_rec = record
    fieldname   : array[1..11] of char;
    fieldtype   : char;                   {C, N, D, L, or M}
    spare1      : integer;  {reserved}
    spare2      : integer;  {reserved}
    width       : byte;     {field length}
    dec         : byte;     {decimal places}
    workspace   : array[1..14] of byte;
  end;


Var
  device        : string[8];
  option        : string[25];
  dbf_file      : file;
  outfile       : text;
  dbf_file_name : string[20];
  outfile_name  : string[20];
  dbf_hdr       : dbf_hdr_rec;
  dbf_fld       : dbf_fld_rec;
  i             : integer;
  fieldcount    : integer;
  names         : array[1..128] of string[10]; {dBase III can have 128 fields/rec}
  lengths       : array[1..128] of byte;
  types         : array[1..128] of char;
  decimals      : array[1..128] of byte;
  line          : integer;
  col           : integer;
  abort         : boolean;
  ch            : char;

{ Command Parsing Variables start here: }

  Type
      parm_array = Array[1..10] Of String[12];
      string255  = String[255];
      string128  = String[128];
      string_2   = String[2];
      string_4   = String[4];

  Var
      ram_tail           : string128 Absolute Cseg : $80;   { MS/PC DOS }
   (* ram_tail           : string128 Absolute $80;          { CP/M 80   } *)
   (* ram_tail           : string128 Absolute Dseg : $80    { CP/M 86   } *)
      blank_string       : string128;
      parm_count         : Integer;
      parms              : parm_array;
      count              : Integer;
      cmd_fname          : String[12];
      mode               : String[12];
      string2            : String[2];

{ Variables for fwrite }
type
  string80             = string[80];

var
  mono                 : boolean;
  cga                  : boolean;
  mono_attr            : Byte;     { character attribute for fwrite - mono }
  cga_attr             : Byte;     { character attribute for fwrite - CGA }
  attr                 : Byte;

{------------------------------------------------------------------------}
{  Procedures Start Here                                                 }
{------------------------------------------------------------------------}
Procedure set_attribute; { set video attribute }
Begin
    cga_attr := 2;   { 0=black, 1=blue, 2=grn, 3=cyan, 4=red, 9=ltblue, 10=ltgrn}
    mono_attr := 11; { 11 = hi nrml, 2 = lo nrml, 1 = lo undln, 9 = hi undln}
    mono := False;
    cga := False;
    { Test for mono or cga }
    If mem[0:$449] = 7 Then mono  := True  { Mono video adapter}
        Else cga  := True;                 { CGA or EGA video adapter}
    If mono Then attr := mono_attr;
    If cga Then attr := cga_attr;   { CGA or EGA }
End;

{-------------------------------------------------------------}
Procedure fwrite(col,row,attrib:byte;str:string80);
    { Fast write routine: 1,1 thru 80,25 }
begin
  col := col - 1;       { col and row are 1 less than the Pascal }
  row := row - 1;       { write routine }

  inline
    ($1E/$1E/$8A/$86/row/$B3/$50/$F6/$E3/$2B/$DB/$8A/$9E/col/
     $03/$C3/$03/$C0/$8B/$F8/$be/$00/$00/$8A/$BE/attrib/
     $8a/$8e/str/$22/$c9/$74/$3e/$2b/$c0/$8E/$D8/$A0/$49/$04/
     $1F/$2C/$07/$74/$22/$BA/$00/$B8/$8E/$DA/$BA/$DA/$03/$46/
     $8a/$9A/str/$EC/$A8/$01/$75/$FB/$FA/$EC/$A8/$01/$74/$FB/
     $89/$1D/$47/$47/$E2/$Ea/$2A/$C0/$74/$10/$BA/$00/$B0/
     $8E/$DA/$46/$8a/$9A/str/$89/$1D/$47/$47/$E2/$F5/$1F);
end;

{------------------------------------------------------------------------}
Procedure center(     row : Integer;
                 the_text : string255);
var
  col   : integer;

begin
  col := (80 - length(the_text)) div 2;
  gotoxy(col,row); write(the_text);
end;

{------------------------------------------------------------------------}
Procedure draw_box(bc,br,ec,er: Integer);
    {begin column, begin row, end column, end row  -  1,1 thru 80,25}
    { set for double line border }
var
  c   : integer;
  r   : integer;

Begin
  if device = 'printer' then exit;
  { draw box }
  c := bc;
  Repeat   { top line }
    fwrite(c,br,attr,Chr(205));  { fwrite(col,row,attrib:byte;str:string80) }
    c := Succ(c);       { next column }
  Until c = ec + 1;
  r := br;
  Repeat   { right side }
    fwrite(ec,r,attr,Chr(186));
    r := Succ(r);       { next row }
  Until r = er + 1;
  c := ec;
  Repeat   { bottom }
    fwrite(c,er,attr,Chr(205));
    c := c - 1;       { next column }
  Until c = bc - 1;
  r := er;
  Repeat   { left side }
    fwrite(bc,r,attr,Chr(186));
    r := r - 1;       { next row }
  Until r = br - 1;
  { 4 corners }  { fwrite(col,row,attrib:byte;str:string80) }
  fwrite(bc,br,attr,Chr(201));  { Upper Left }
  fwrite(ec,br,attr,Chr(187));  { Upper Right }
  fwrite(ec,er,attr,Chr(188));  { Lower Right }
  fwrite(bc,er,attr,Chr(200));  { Lower Left }
End;

{------------------------------------------------------------------------}
procedure signon;
begin
  clrscr;
  draw_box(1,2,80,6);    {(l,t,r,b: Integer)}
  gotoxy(3,1); writeln(' Program - RDBF.PAS ');
  gotoxy(54,1); writeln('Tom Woodgerd - 11/02/90 ');
end;

{------------------------------------------------------------------------}
procedure show_options;
begin
  draw_box(1,7,80,17);    {(l,t,r,b: Integer)}
  gotoxy(3,08); writeln('        COMMAND LINE OPTIONS');
  gotoxy(3,09); writeln('RDBF             - Prompts for file name.');
  gotoxy(3,10); writeln('RDBF filename    - Displays the file definition.');
  gotoxy(3,11); writeln('RDBF filename f  - Displays and creates "filename.DEF".');
  gotoxy(3,12); writeln('RDBF filename p  - Displays and prints the file definition.');
  gotoxy(3,14); writeln('The file extension (.DBF) is optional.')
end;

{---------------------------------------------------------------------------}
Function strip_white(target : string255) : string255;
Const whitespace : Set of Char = [#8,#10,#12,#13,' '];

Begin
    While (Length(target) > 0) And (target[1] In whitespace)
        Do Delete(target,1,1);
    strip_white := target;
End;

{---------------------------------------------------------------------------}
Procedure read_parms;   { read command line parameters }
                        { return with: parm_count = # of parms }
                        {              parms[I]   = array 1-10 of strings }
Var
    marker    : Integer;
    tail      : String[128]; { allow a 128 byte command tail }

Begin
    tail := ram_tail;        { ram_tail contains the command tail }
    parms[1] := '           ';
    parms[2] := '           ';
    parms[3] := '  ';
    count := 0;
    tail := strip_white(tail);       { use function to strip white chars }
    If Length(tail) > 0 Then Repeat
        marker := Pos(' ',tail) - 1;
        If marker <= 0 Then marker := Length(tail);
        count := count + 1;
        parms[count] := Copy(tail,1,marker);
        Delete(tail,1,marker);
        tail := strip_white(tail);
    Until (count >= 10) Or (Length(tail) = 0);
    parm_count := count;
End;

{----------------------------------------------------------------------------}
Procedure get_name;

Begin
  center(4,'Input .DBF File Name: ');
  Read(dbf_file_name);
End;

{----------------------------------------------------------------------------}
Procedure open_dbf_file;
Begin
  { add the filetype if necessary}
  i := Pos('.',dbf_file_name);
  If i = 0 Then dbf_file_name := dbf_file_name + '.DBF'; { add filetype  }
  assign(dbf_file,dbf_file_name);
  gotoxy(5,6); writeln('  Opening:  ',dbf_file_name,'  ');
  {$I-} reset(dbf_file,1); {$I+}
  i := Ioresult;
  If i <> 0 Then Begin
    gotoxy(5,20);
    Writeln('    ',dbf_file_name,' Not Found! ');
    Halt
  End;
  seek(dbf_file,0);    {point to first byte}
  clrscr;
  signon;
  for i := 1 to length(dbf_file_name) do begin
    dbf_file_name[i] := upcase(dbf_file_name[i]);
  end;
  gotoxy(10,3); write('.DBF File  : ',dbf_file_name);
End;

{------------------------------------------------------------------------}
procedure get_hdr;
const
  c65536 = 65536.0;

var
  real_count  : real;
  real_count1 : real;
  real_count2 : real;

begin
  real_count := 0;
  seek(dbf_file,0);    {point to first byte}
  blockread(dbf_file,dbf_hdr,sizeof(dbf_hdr));
  { Do record count conversions }
  with dbf_hdr do begin
    real_count1 := rec_count1;
    real_count2 := rec_count2;
  end;
  if real_count1 < 0 then       { Convert integer to unsigned integer (real) }
    real_count1 := (c65536 + real_count1);
  if real_count2 < 0 then       { Convert integer to unsigned integer (real) }
    real_count2 := (c65536 + real_count2);
  real_count := (c65536 * real_count2) + real_count1;
  if device = 'screen' then begin
    with dbf_hdr do begin
      gotoxy(10,4); write('File Type  : ',headtype,'  ');
      gotoxy(28,4);
      case headtype of
        $03 : write('(No Memo File)');
        $83 : write('(Memo File Exists)');
      else
        write('Not a dBase File!');
        abort := true;
      end;
      if abort = false then begin
        gotoxy(10,5); write('Last Update: ',month,'/',day,'/',year);
        gotoxy(50,3); write('Total Records: ',real_count:7:0);
        gotoxy(50,4); write('Header Length:     ',hdr_length);
        gotoxy(50,5); write('Record Size  :     ',rec_size);
      end;
    end;
  end;

  if device = 'printer' then begin
    writeln(lst,'=========================================');
    writeln(lst,'     ','.DBF File  : ',dbf_file_name);
    writeln(lst,'=========================================');
    with dbf_hdr do begin
      write(lst,'     ','File Type    : ',headtype,'  ');
      case headtype of
        $03 : writeln(lst,'(No Memo File)');
        $83 : writeln(lst,'(Memo File Exists)');
      else
        writeln(lst,'NOT A dBASE FILE!');
        abort := true;
      end;
      writeln(lst,'     ','Last Update  : ',month,'/',day,'/',year);
      writeln(lst,'     ','Total Records: ',real_count:7:0);
      writeln(lst,'     ','Header Length:     ',hdr_length);
      writeln(lst,'     ','Record Size  :     ',rec_size);
    end;
    writeln(lst,'');
  end;

  if device = 'file' then begin
    writeln(outfile,'=========================================');
    writeln(outfile,'     ','.DBF File  : ',dbf_file_name);
    writeln(outfile,'=========================================');
    with dbf_hdr do begin
      write(outfile,'     ','File Type    : ',headtype,'  ');
      case headtype of
        $03 : writeln(outfile,'(No Memo File)');
        $83 : writeln(outfile,'(Memo File Exists)');
      else
        writeln(outfile,'NOT A dBASE FILE!');
        abort := true;
      end;
      writeln(outfile,'     ','Last Update  : ',month,'/',day,'/',year);
      writeln(outfile,'     ','Total Records: ',real_count:7:0);
      writeln(outfile,'     ','Header Length:     ',hdr_length);
      writeln(outfile,'     ','Record Size  :     ',rec_size);
    end;
    writeln(outfile,'');
  end;
end;

{------------------------------------------------------------------------}
procedure processfield(fld : dbf_fld_rec;
                       i   : integer);
var
  j   : integer;

begin
  with dbf_fld do begin
    names[i] := '';
    j := 1;                {char count}
    while (j < 11) and (fieldname[j] <> #0) do begin; {#0 = end of name}
      names[i] := names[i] + fieldname[j];
      j := j + 1;
    end;
    types[i] := fieldtype;
    lengths[i] := width;
    decimals[i] := dec;
    if device = 'screen' then begin
      gotoxy(col+7,line); write(i);
      gotoxy(col+10,line); write(names[i]);
      gotoxy(col+24,line); write(types[i]);
      gotoxy(col+27,line); write(lengths[i]:3);
      if types[i] = 'N' then begin
        gotoxy(col+33,line); writeln(decimals[i]:2);
      end;
    end;
    if device = 'printer' then begin
      write(lst,'     ',i:3,'  ');
      write(lst,names[i]);
      for j := 1 to (12 - length(names[i])) do begin
        write(lst,' ');
      end;
      write(lst,types[i]:4);
      for j := 1 to length(types[i]) - 6 do begin write(lst,' '); end;
      write(lst,lengths[i]:4);
      write(lst,'   ');
      if types[i] = 'N' then begin
        write(lst,decimals[i]:2);
      end;
      writeln(lst,'');
    end;
    if device = 'file' then begin
      write(outfile,'     ',i:3,'  ');
      write(outfile,names[i]);
      for j := 1 to (12 - length(names[i])) do begin
        write(outfile,' ');
      end;
      write(outfile,types[i]:4);
      for j := 1 to length(types[i]) - 6 do begin write(outfile,' '); end;
      write(outfile,lengths[i]:4);
      write(outfile,'   ');
      if types[i] = 'N' then begin
        write(outfile,decimals[i]:2);
      end;
      writeln(outfile,'');
    end;
  end;
end;

{------------------------------------------------------------------------}
procedure get_fld_def;
var
  kbchar : char;

begin
  if device = 'printer' then begin
    writeln(lst,'                   FIELDS');
    writeln(lst);
  end;
  if device = 'file' then begin
    writeln(outfile,'                   FIELDS');
    writeln(outfile);
  end;
  gotoxy(31,7); writeln('Field Definitions');
  line := 9;
  col := 0;
  seek(dbf_file,32);  {this is actually the 33rd byte}
  i := 1;
  dbf_fld.fieldname[1] := ' ';
  while (dbf_fld.fieldname[1] <> chr(13)) do begin
    {read one byte to avoid eof err when no records exist}
    blockread(dbf_file,dbf_fld.fieldname[1],1);
    if dbf_fld.fieldname[1] <> chr(13) then begin  {check for end of fields}
      blockread(dbf_file, dbf_fld.fieldname[2], sizeof(dbf_fld) - 1);
      processfield(dbf_fld,i);
      line := line + 1;
      if line > 24 then begin
        if col = 40 then begin
          if device = 'screen' then begin
            center(25,'Press Any Key For More Fields');
            read(kbd,kbchar);
          end;
          for line := 9 to 24 do begin
            gotoxy(1,line);
            clreol;
          end;
          col := 0;
          line := 9;
        end
        else begin
          col := 40;
          line := 9;
        end;
      end;
      i := i + 1;
    end;
  end;
  fieldcount := i - 1;
end;

{------------------------------------------------------------------------}
procedure print_it;
begin
  device := 'printer';
  get_hdr;
  get_fld_def;
  {write(lst,chr(12));}  {formfeed}
  writeln(lst); writeln(lst);  writeln(lst);
  device := 'screen';
end;

{------------------------------------------------------------------------}
procedure file_it;
begin
  i := Pos('.',dbf_file_name);
  outfile_name := copy(dbf_file_name,1,i-1);
  outfile_name := outfile_name + '.DEF';
  assign(outfile,outfile_name);
  gotoxy(1,25);
  write('               Creating: ',outfile_name,'                             ');
  {$I-} rewrite(outfile); {$I+}
  i := Ioresult;
  If i <> 0 Then Begin
    gotoxy(1,25);
    Write('               Unable to create ',outfile_name,'                      ');
    Halt
  End;
  device := 'file';
  get_hdr;
  get_fld_def;
  device := 'screen';
  close(outfile);
end;

{------------------------------------------------------------------------}
{ Main                                                                   }
{------------------------------------------------------------------------}
begin
  device := 'screen';
  abort := false;
  set_attribute;
  signon;   { show logo }
  read_parms;    { get filenames from command line }
  If parm_count < 1 Then begin
    show_options;
    get_name;    { prompt for filenames }
  end
  else begin
    dbf_file_name := parms[1];
  end;
  open_dbf_file;
  get_hdr;
  get_fld_def;
  gotoxy(1,25);
  if parm_count > 1 then begin
    option := parms[2];
  end
  else begin
    write(' Send to <P>rinter or <F>ile or Enter to quit. <P, F, or Enter> ');
    read(option);
  end;
  if (option = 'P') or (option = 'p') then print_it;
  if (option = 'F') or (option = 'f') then file_it;
  gotoxy(1,24);
end.
